#ifdef HAVE_CONFIG_H
#  include "../config.h"
#endif

#include <string.h>

#include <beryl.h>

#define DETECT_REFRESH_RATE_DEFAULT TRUE

#define SCREEN_SIZE_DEFAULT  4
#define SCREEN_SIZE_MIN	     2
#define SCREEN_SIZE_MAX	     32

#define SCREEN_VSIZE_DEFAULT 1
#define SCREEN_VSIZE_MIN     1
#define SCREEN_VSIZE_MAX     32

#define LIGHTING_DEFAULT TRUE

#define OPACITY_STEP_DEFAULT 10
#define OPACITY_STEP_MIN     1
#define OPACITY_STEP_MAX     50

#define UNREDIRECT_FS_DEFAULT FALSE

#define DEFAULT_ICON_DEFAULT "icon.png"

#define SYNC_TO_VBLANK_DEFAULT TRUE

#define DEFAULT_REFRESH_RATE 60

#ifndef NUM_OPTIONS
#define NUM_OPTIONS(s) (sizeof ((s)->opt) / sizeof (CompOption))
#endif

void
compScreenInitOptions (CompScreen * screen);

void
compScreenInitOptions (CompScreen * screen)
{
    CompOption *o;

    o = &screen->opt[COMP_SCREEN_OPTION_DETECT_REFRESH_RATE];
    o->name = "detect_refresh_rate";
    o->shortDesc = N_("Detect Refresh Rate");
    o->longDesc = N_("Automatic detection of refresh rate");
    o->type = CompOptionTypeBool;
    o->value.b = DETECT_REFRESH_RATE_DEFAULT;

    o = &screen->opt[COMP_SCREEN_OPTION_LIGHTING];
    o->name = "lighting";
    o->shortDesc = N_("Lighting");
    o->longDesc = N_("Use diffuse light when screen is transformed");
    o->type = CompOptionTypeBool;
    o->value.b = LIGHTING_DEFAULT;

    o = &screen->opt[COMP_SCREEN_OPTION_REFRESH_RATE];
    o->name = "refresh_rate";
    o->shortDesc = N_("Refresh Rate");
    o->longDesc =
        N_("The rate at which the screen is redrawn (times/second)");
    o->type = CompOptionTypeInt;
    o->value.i = DEFAULT_REFRESH_RATE;
    o->rest.i.min = 1;
    o->rest.i.max = 200;

    o = &screen->opt[COMP_SCREEN_OPTION_SIZE];
    o->name = "size";
    o->shortDesc = N_("Horizontal Virtual Size");
    o->longDesc = N_("Screen size multiplier for virtual size");
    o->type = CompOptionTypeInt;
    o->value.i = SCREEN_SIZE_DEFAULT;
    o->rest.i.min = SCREEN_SIZE_MIN;
    o->rest.i.max = SCREEN_SIZE_MAX;

    o = &(screen->opt[COMP_SCREEN_OPTION_VSIZE]);
    o->name = "vsize";
    o->shortDesc = N_("Vertical Virtual Size");
    o->longDesc = N_("Screen size multiplier for vertical virtual size");
    o->type = CompOptionTypeInt;
    o->value.i = SCREEN_VSIZE_DEFAULT;
    o->rest.i.min = SCREEN_VSIZE_MIN;
    o->rest.i.max = SCREEN_VSIZE_MAX;

    o = &screen->opt[COMP_SCREEN_OPTION_OPACITY_STEP];
    o->name = "opacity_step";
    o->shortDesc = N_("Opacity Step");
    o->longDesc = N_("Opacity change step");
    o->type = CompOptionTypeInt;
    o->value.i = OPACITY_STEP_DEFAULT;
    o->rest.i.min = OPACITY_STEP_MIN;
    o->rest.i.max = OPACITY_STEP_MAX;

    o = &screen->opt[COMP_SCREEN_OPTION_UNREDIRECT_FS];
    o->name = "unredirect_fullscreen_windows";
    o->shortDesc = N_("Unredirect Fullscreen Windows");
    o->longDesc =
        N_("Allow drawing of fullscreen windows to not be redirected "
           "to offscreen pixmaps");
    o->type = CompOptionTypeBool;
    o->value.b = UNREDIRECT_FS_DEFAULT;

    o = &screen->opt[COMP_SCREEN_OPTION_DEFAULT_ICON];
    o->name = "default_icon";
    o->shortDesc = N_("Default Icon");
    o->longDesc = N_("Default window icon image");
    o->type = CompOptionTypeString;
    o->value.s = strdup (DEFAULT_ICON_DEFAULT);
    o->rest.s.string = 0;
    o->rest.s.nString = 0;

    o = &screen->opt[COMP_SCREEN_OPTION_SLOWNESS_FIX];
    o->name = "slowness_fix";
    o->shortDesc = N_("Slowness Fix");
    o->longDesc = N_("Toggle this option if beryl is slow or choppy. "
                     "On some cards, enabling it makes beryl smoother, and on "
                     "others, disabling makes beryl smoother.");
    o->type = CompOptionTypeBool;
    o->value.b = TRUE;

    o = &screen->opt[COMP_SCREEN_OPTION_SYNC_TO_VBLANK];
    o->name = "sync_to_vblank";
    o->shortDesc = N_("Sync To VBlank");
    o->longDesc = N_("Only perform screen updates during vertical "
                     "blanking period");
    o->type = CompOptionTypeBool;
    o->value.b = SYNC_TO_VBLANK_DEFAULT;
}

CompOption *
compGetScreenOptions (CompScreen * screen, int *count)
{
    *count = NUM_OPTIONS (screen);
    return screen->opt;
}

CompOption *
compGetOptions (int *count)
{
    static CompScreen screen;
    compScreenInitOptions(&screen);
    return compGetScreenOptions(&screen, count);
}


